"""Reading and writing to files.
"""

from __future__ import annotations

import os
from typing import Any


def __export_to_tempfile(code: str) -> str:
    """Export latex code to temp file"""
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        print("Writing to tempfile")
        f.write(code)
        # f.flush()
    return f.name


def export_pdf_from_code(code: str) -> str:
    """
    Compile the given ``tex`` code to a pdf file.

    Use ``pdflatex`` to compile the document to a standalone pdf file.

    Parameters
    ----------
    code
        String representation of the tex to be compiled.

    Returns
    -------
    str
        Path to the generated pdf file.
    """
    file = __export_to_tempfile(code)
    return export_pdf_from_file(file)


def export_pdf_from_file(path: str) -> str:
    """
    Compile the ``tex`` code at the given path.


    Use ``pdflatex`` to compile the document to a standalone pdf file.

    Parameters
    ----------
    path
        The path to the ``tex`` code to be compiled.

    Returns
    -------
    str
        The path to the generated pdf file.

    Raises
    ------
    PDFlatexNotFoundError:
        If pdflatex is not installed.
    CompilationError:
        If the given document could not be compiled by pdflatex.
    """
    import subprocess
    from .exceptions import PDFlatexNotFoundError, CompilationError
    import os
    working_dir = os.path.dirname(path)

    options: dict[str, Any] = dict(capture_output=True, check=True)
    if working_dir:
        options["cwd"] = working_dir

    try:
        import shutil
        pdflatex_path = shutil.which('pdflatex')
        if pdflatex_path is None:
            raise PDFlatexNotFoundError(f"Could not find executable `pdflatex` to compile {path}. Please make sure it is installed and accessible in the system's path.")

        subprocess.run([pdflatex_path, "-interaction=nonstopmode", "-halt-on-error", path], **options)
        # print("Pdflatex done!")
    except subprocess.CalledProcessError as e:
        error_message = f"{e.stdout.decode('UTF-8')}\n\nCompilation failed with the error above ☝️ "
        raise CompilationError(error_message)
    import os

    basename = os.path.splitext(path)[0]
    for ext in {".aux", ".log"}:
        try:
            os.remove(basename + ext)
        except FileNotFoundError:
            ...
    return f"{basename}.pdf"


def export_png_from_file(input_file: str, **options) -> str:
    """
    Export the given tex file to a png image.

    The tex file is first compiled to pdf using pdflatex. Then the
    resulting pdf is converted to an image using ``pdf2image``.

    Parameters
    ----------
    input_file
        The path to the ``tex`` code to be compiled.

    Returns
    -------
    str
        The path to the generated png file.
    """
    import pdf2image
    pdf_file = export_pdf_from_file(input_file)
    path = pdf_file.replace(".pdf", ".png")
    options["output_file"] = path
    pdf2image.convert_from_path(pdf_file, **options)
    return path


def export_png_from_code(code: str, path: str, **options):
    """
    Export the given tex file to a png image.

    The tex file is first compiled to pdf using pdflatex. Then the
    resulting pdf is converted to an image using ``pdf2image``.

    Parameters
    ----------
    input_file
        The path to the ``tex`` code to be compiled.

    Returns
    -------
    str
        The path to the generated png file.
    """
    import pdf2image
    pdf_file = export_pdf_from_code(code)
    options["output_file"] = path
    pdf2image.convert_from_path(pdf_file, **options)


def open_pdf_file(file_path: str):
    """
    Open the pdf file at the given path in the system default pdf reader.

    Parameters
    ----------
    file_path
        Path to the pdf file to open.
    """
    import sys
    import subprocess
    print("Opening pdf file")

    if sys.platform.startswith('darwin'):  # macOS
        result = subprocess.run(['open', file_path], capture_output=True, text=True)
    elif sys.platform.startswith('win32'):  # Windows
        import os
        result = os.startfile(file_path)
    elif sys.platform.startswith('linux'):  # Linux
        result = subprocess.run(['xdg-open', file_path], capture_output=True, text=True)
    else:
        print("Unsupported platform. Unable to open PDF file.")
    input("Press any key to continue.")
    if result.returncode != 0:
        print(sys.stdout)


def preview_latex_doc(code: str) -> str:
    """
    Preview the given tex code.

    The given code must be a valid tex document that can be compiled.
    It gets compiled to a temporary file and then opened in the default pdf reader.

    Parameters
    ----------
    code
        The tex to be compiled.

    Returns
    -------
    str
        The path to which the compiled pdf was written. Could be useful for cleanup.
    """
    pdf_path = export_pdf_from_code(code)
    print(pdf_path)

    try:
        open_pdf_file(pdf_path)
    except Exception as e:
        try:
            os.remove(pdf_path)
        except FileNotFoundError():
            ...
        raise e


if __name__ == "__main__":

    TEX_STRING = r"""
    \documentclass{article}
    \begin{document}
        hello world.
    \end{document}
    """
    preview_latex_doc(TEX_STRING)



"""Public API"""

from __future__ import annotations

from .environments.axis import Axis
from .environments.tikzpicture import TikzPicture
from .commands.addplot import Addplot
from .commands.fillbetween import FillBetween
from .commands.node import Node
from .commands.draw import Draw
from .commands.circle import Circle
from .plot import create_plot
import numpy as np
from typing import Optional, Union


class WorkSpace:
    __currfig = None
    __currax = None

    @classmethod
    def scf(cls, fig: TikzPicture):
        cls.__currfig = fig

    @classmethod
    def sca(cls, ax: TikzPicture):
        cls.__currax = ax

    @classmethod
    def gcf(cls) -> Optional[Axis]:
        return cls.__currfig

    @classmethod
    def gca(cls) -> Optional[Axis]:
        return cls.__currax


def gcf() -> Optional[TikzPicture]:
    """
    Get the currently active figure in the workspace, which could be None.
    If you want to create a new one if there is no current figure, then use `figure()`

    Returns
    -------
    Optional[TikzPicture]
        The currently active Figure if there is one.
    """
    return WorkSpace.gcf()


def gca() -> Optional[Axis]:
    """
    Get the currently active axis (representing a pgfplots ``axis``) if there is one.

    If you want to create a new axis in case there is none, then use the `ax()` function.

    Returns
    -------
    Optional[Axis]
        The currently active axis if there is one.
    """
    return WorkSpace.gca()


def __get_or_create_fig() -> TikzPicture:
    f = gcf()
    return f if f is not None else figure()


def __get_or_create_ax() -> Axis:
    a = gca()
    return a if a is not None else ax()


def figure(standalone: bool = True, **options) -> TikzPicture:
    """
    Create a new TikzPicture and set it as the current figure.

    Parameters
    ----------
    standalone : bool, optional
        Whether the figure should be a standalone document, by default True
    **options
        Additional options to pass to the TikzPicture

    Returns
    -------
    TikzPicture
        The newly created figure
    """
    pic = TikzPicture(standalone, **options)
    WorkSpace.sca(None)
    WorkSpace.scf(pic)
    return pic


def ax(**options) -> Axis:
    """
    Create a new Axis object and add it to the current figure.

    Parameters
    ----------
    **options
        Options to pass to the Axis constructor

    Returns
    -------
    Axis
        The newly created axis
    """
    ax = Axis(**options)
    fig = __get_or_create_fig()
    WorkSpace.sca(ax)
    fig.add_axis(ax)
    return ax


def dumps(fig: TikzPicture | None = None) -> str:
    """
    Dump the tex-code of the current figure to a string.

    If no figure is given, the active figure is selected,
    and the contents of this figure are returned.

    Parameters
    ----------
    fig
        Figure to generate the code for. If None is given, then the current figure is used.

    Returns
    -------
    str
        The tex code.
    """
    fig = gcf() if fig is None else fig
    if fig is None:
        return ""
    return fig.get_code()


def preview(fig: TikzPicture | None = None):
    """
    Preview the current figure.

    The contents of this figure are written to a temporary file,
    compiled to a pdf and then opened in the default viewer.
    After the process terminates, these files are removed.

    Parameters
    ----------
    fig : TikzPicture, optional
        The figure to preview. If None is given, the current figure is used.
    """
    fig = gcf() if fig is None else fig
    if fig is None:
        return
    fig.preview()


def save(filename: str, fig: TikzPicture | None = None, standalone: bool = False):
    """
    Save the generated Tikz code to a file.

    Parameters
    ----------
    filename : str
        The path to write the figure to.
        If the extension `.tex` or `.tikz` is not present, `.tex` is appended.
    fig : TikzPicture, optional
        The figure to save. If None is given, the current figure is used.
    standalone : bool, optional
        Whether to save as a standalone document, by default False
    """

    fig = gcf() if fig is None else fig
    if fig is None:
        return
    fig.standalone = standalone
    return fig.export(filename)


def xlabel(lab: str):
    """
    Set the x label of the current axis.

    Parameters
    ----------
    lab : str
        Label text
    """
    ax = __get_or_create_ax()
    ax.set_xlabel(lab)


def xticks(ticks: np.ndarray, labels: list[str] | None = None, ax: Axis | None = None):
    """
    Set the x ticks and optionally label them.

    Parameters
    ----------
    ticks : np.ndarray
        x-coordinates to add ticks at
    labels : list[str], optional
        Labels of the ticks
    ax : Axis, optional
        The axis to modify. The current axes is used if None is passed.
    """
    ax = __get_or_create_ax() if ax is None else ax
    ax.set_xticks(ticks, labels)


def yticks(ticks: np.ndarray, labels: list[str] | None = None, ax: Axis | None = None):
    """
    Set the y ticks and optionally label them.

    Parameters
    ----------
    ticks : np.ndarray
        y-coordinates to add ticks at
    labels : list[str], optional
        Labels of the ticks
    ax : Axis, optional
        The axis to modify. The current axes is used if None is passed.
    """
    ax = __get_or_create_ax() if ax is None else ax
    ax.set_yticks(ticks, labels)


def zticks(ticks: np.ndarray, labels: list[str] | None = None, ax: Axis | None = None):
    """
    Set the z ticks and optionally label them.

    Parameters
    ----------
    ticks : np.ndarray
        z-coordinates to add ticks at
    labels : list[str], optional
        Labels of the ticks
    ax : Axis, optional
        The axis to modify. The current axes is used if None is passed.
    """
    ax = __get_or_create_ax() if ax is None else ax
    ax.set_zticks(ticks, labels)


def define_style(name: str, fig: TikzPicture | None = None, **options):
    """
    Define a new style for the figure.

    Parameters
    ----------
    name : str
        Name of the style to define
    fig : TikzPicture, optional
        Figure to define the style for. If None, uses current figure.
    **options
        Style options to set
    """
    fig = __get_or_create_fig() if fig is None else fig
    fig.set_style(name, **options)


def ylabel(lab: str):
    """
    Set the y label of the current axis.

    Parameters
    ----------
    lab : str
        Label text
    """
    ax = __get_or_create_ax()
    ax.set_ylabel(lab)


def zlabel(lab: str):
    """
    Set the z label of the current axis.

    Parameters
    ----------
    lab : str
        Label text
    """
    ax = __get_or_create_ax()
    ax.set_zlabel(lab)


def xlim(limits: tuple[float]):
    """
    Set the x limits of the current axis.

    Parameters
    ----------
    limits : tuple[float]
        Tuple containing (min, max) limits
    """
    ax = __get_or_create_ax()
    ax.set_xlims(limits)


def ylim(limits: tuple[float]):
    """
    Set the y limits of the current axis.

    Parameters
    ----------
    limits : tuple[float]
        Tuple containing (min, max) limits
    """
    ax = __get_or_create_ax()
    ax.set_ylims(limits)


def axhline(y: float, ax: Axis | None = None, **options) -> list[Addplot]:
    """
    Plot a horizontal line at the provided y value.

    Parameters
    ----------
    y : float
        y-coordinate of the line
    ax : Axis, optional
        Axis to add the line to. If None, uses current axis.
    **options
        Additional plotting options

    Returns
    -------
    list[Addplot]
        List of plot commands created
    """
    return plot(y, ax, **options)


def scale(scale: float, fig: TikzPicture | None = None):
    """
    Set the scale of the figure.

    Parameters
    ----------
    scale : float
        Scale factor
    fig : TikzPicture, optional
        Figure to scale. If None, uses current figure.
    """
    fig = gcf() if fig is None else fig
    fig.set_option("scale", scale)


def point(coordinates: np.ndarray | None = None,
          label: str = "",
          name: str | None = None,
          axis_coords: bool | None = None,
          label_loc: str = "above",
          size: int | str = "1pt",
          axis: Axis | None = None,
          fig: TikzPicture | None = None,
          **options) -> Node:
    """
    Draw a point at the given coordinates.

    It creates a node, but also adds some styling to make it look like a point.

    See also: node()

    Parameters
    ----------
    coordinates : np.ndarray, optional
        Position of the point
    label : str, optional
        Label text, by default ""
    name : str, optional
        Name of the node
    axis_coords : bool, optional
        Whether to use axis coordinates
    label_loc : str, optional
        Label location ("above", "below", "right", "left"), by default "above"
    size : int | str, optional
        Size of the point, by default "1pt"
    axis : Axis, optional
        Axis to add the point to
    fig : TikzPicture, optional
        Figure to add the point to
    **options
        Additional styling options

    Returns
    -------
    Node
        The created node object
    """
    options["pt"] = True
    nd = node(coordinates, label, name, axis_coords, label_loc, axis, fig, **options)
    fig = fig if fig is not None else __get_or_create_fig()
    fig.define_style("pt", inner_sep=size, circle=True, fill=True)
    return nd


def node(coordinates: np.ndarray | None = None,
         label: str | None = None,
         name: str | None = None,
         axis_coords: bool | None = None,
         label_loc: str | None = None,
         axis: Axis | None = None,
         fig: TikzPicture | None = None,
         **options) -> Node:
    """
    Draw a node at the given coordinates.

    Parameters
    ----------
    coordinates : np.ndarray, optional
        Position of the node or name of a previously defined node
    label : str, optional
        Label text
    name : str, optional
        Name of the node
    axis_coords : bool, optional
        Whether to interpret coordinates in axis coordinates
    label_loc : str, optional
        Label location (None, "above", "below", "right", "left" or combinations)
    axis : Axis, optional
        Axis to add the node to
    fig : TikzPicture, optional
        Figure to add the node to
    **options
        Additional styling options

    Returns
    -------
    Node
        The created node object
    """
    if axis_coords is None:
        given_axis = gca() if axis is None else axis
        axis_coords = False if given_axis is None else True

    if not label:
        label_loc = None
    node = Node(label, name, coordinates, label_loc, axis_coords, **options)
    if axis is None and axis_coords:
        axis = __get_or_create_ax()
        axis.add(node)
        return node

    fig = fig if fig is not None else __get_or_create_fig()
    fig.add(node)
    return node


def fill_between(x: np.ndarray, y1: np.ndarray, y2: np.ndarray, *,
                 draw_options=None, fill_options=None) -> FillBetween:
    """
    Fill the area between two curves.

    Parameters
    ----------
    x : np.ndarray
        x-coordinates
    y1 : np.ndarray
        y-coordinates of first curve
    y2 : np.ndarray
        y-coordinates of second curve
    draw_options : dict, optional
        Options for drawing the curves
    fill_options : dict, optional
        Options for filling the area

    Returns
    -------
    FillBetween
        The fill between plot command
    """
    ax = __get_or_create_ax()
    ax.add_requirement(r"\usepgfplotslibrary{fillbetween}")

    draw_options = draw_options if draw_options is not None else dict()
    fill_options = fill_options if fill_options is not None else dict()

    draw_options["name path"] = "first"
    plot(x, y1, ax, **draw_options)

    draw_options["name path"] = "second"
    plot(x, y2, ax, **draw_options)[0]

    fill_plot = FillBetween("first", "second", **fill_options)
    ax.add(fill_plot)
    return fill_plot


# def quiver(x: np.ndarray, y: np.ndarray, dx: np.ndarray, dy: np.ndarray, **options) -> Quiver


def axvline(x: float, ax: Axis | None = None, **options) -> list[Addplot]:
    """
    Plot a vertical line at the provided x value.

    Parameters
    ----------
    x : float
        x-coordinate of the line
    ax : Axis, optional
        Axis to add the line to. If None, uses current axis.
    **options
        Additional plotting options

    Returns
    -------
    list[Addplot]
        List of plot commands created
    """    # TODO: Think of ways to represent this allow axes dimensions to be updated afterwards.
    ax = ax if ax is not None else __get_or_create_ax()
    ymin, ymax = ax.get_ylims()
    xes = np.array((x, x))
    ys = np.array((ymin, ymax))
    return plot(xes, ys, ax=ax, **options)


Point = Union[np.ndarray, str, Node]


def __create_draw(connector_type: str, points: list[Point], **options) -> Draw:
    from .commands import Connector
    connector = Connector(connector_type)
    draw = Draw(points, connector, **options)
    return draw


def __add_draw_command(draw: Draw) -> Draw:
    fig = __get_or_create_fig()
    fig.add(draw)
    return draw


def __create_and_add_draw(connector_type: str, points: list[Point], **options):

    draw = __create_draw(connector_type, points, **options)
    return __add_draw_command(draw)


def rectangle(c1: Point, c2: Point, **options) -> Draw:
    """
    Draw a rectangle between two points.

    Parameters
    ----------
    c1 : Point
        First corner point
    c2 : Point
        Second corner point
    **options
        Additional drawing options

    Returns
    -------
    Draw
        The drawing command
    """
    return __create_and_add_draw("rectangle", [c1, c2], **options)


def circle(center: Point, radius: float, **options) -> Circle:
    """
    Draw a circle with given center and radius.

    Parameters
    ----------
    center : Point
        Center point of the circle
    radius : float
        Radius of the circle
    **options
        Additional drawing options

    Returns
    -------
    Circle
        The circle drawing command
    """
    circle = Circle(center, radius, **options)
    return __add_draw_command(circle)


def line(points: list[Point], connection: str = "--", **options) -> Draw:
    """
    Draw a line through a list of points.

    Parameters
    ----------
    points : list[Point]
        List of points to connect
    connection : str, optional
        Type of connection between points, by default "--"
    **options
        Additional drawing options

    Returns
    -------
    Draw
        The drawing command
    """
    return __create_and_add_draw(connection, points, **options)


def arrow(points: list[Point], forward: bool = True, backward: bool = False,
          arrowhead: str | None = None, **options) -> Draw:
    """
    Draw an arrow through a list of points.

    Parameters
    ----------
    points : list[Point]
        List of points to connect
    forward : bool, optional
        Whether to draw forward arrow, by default True
    backward : bool, optional
        Whether to draw backward arrow, by default False
    arrowhead : str, optional
        Style of arrowhead
    **options
        Additional drawing options

    Returns
    -------
    Draw
        The drawing command
    """
    bw = "<" if backward else ""
    fw = ">" if forward else ""
    arrow = bw + "-" + fw if (bw or fw) else "--"
    arrow_opts = {arrow: True}
    if arrowhead is not None:
        if forward:
            arrow_opts[">"] = arrowhead
        if backward:
            arrow_opts["<"] = arrowhead
    options.update(arrow_opts)
    return line(points, **options)


def plot(x, y=None, z=None, ax: Axis | None = None, label: str | tuple[str] | None = None,
         inline_label: bool = False,
         **options) -> list[Addplot]:
    """
    Create a plot command.

    Parameters
    ----------
    x : array-like
        x-coordinates or data
    y : array-like, optional
        y-coordinates
    z : array-like, optional
        z-coordinates for 3D plots
    ax : Axis, optional
        Axis to add the plot to
    label : str | tuple[str], optional
        Label or tuple of labels for the plot
    inline_label : bool, optional
        Whether to place label inline with the plot, by default False
    **options
        Additional plotting options

    Returns
    -------
    list[Addplot]
        List of plot commands created
    """

    ax = ax if ax is not None else __get_or_create_ax()
    plot_commands = create_plot(x, y, z, label, inline_label, **options)
    for plt in plot_commands:
        ax.add(plt)
    return plot_commands


def scatter(x, y=None, z=None, ax: Axis | None = None, label: str | tuple[str] | None = None,
            inline_label: bool = False,
            **options) -> list[Addplot]:
    """
    Create a scatter plot command.

    Parameters
    ----------
    x : array-like
        x-coordinates or data
    y : array-like, optional
        y-coordinates
    z : array-like, optional
        z-coordinates for 3D plots
    ax : Axis, optional
        Axis to add the plot to
    label : str | tuple[str], optional
        Label or tuple of labels for the plot
    inline_label : bool, optional
        Whether to place label inline with the plot, by default False
    **options
        Additional plotting options

    Returns
    -------
    list[Addplot]
        List of plot commands created
    """
    options["only marks"] = True
    return plot(x, y, z, ax, label, inline_label, **options)



from __future__ import annotations

from .tikzcode import Tex
from .options import OptionsMixin


class Command(Tex, OptionsMixin):
    r"""A LaTeX Command of the form `\<cmdname>[<options>]<middle>{<arg1>}{<arg2>}...`."""

    def __init__(self, cmd: str, *arguments, **options):
        self.cmd_name: str = cmd
        # self._options: Options = Options(**options)
        self._arguments: list[Tex] = []
        for argument in arguments:
            self.add_argument(argument)
        self.init_options(**options)
        self._endline = True
        self._escape = True

    def add_argument(self, argument: str | Tex):
        if isinstance(argument, str):
            argument = Tex(argument)
        self._arguments.append(argument)

    def _format_middle(self) -> str:
        return ""

    def _format_post(self) -> str:
        return ""

    @property
    def arguments(self) -> list[Tex]:
        return self._arguments

    def _format_arguments(self) -> str:
        argument_str = ""
        for argument in self.arguments:
            argument_str += f"{{{argument.get_code()}}}"
        return argument_str

    def get_code(self) -> str:
        options_str = self.options.format()
        argument_str = self._format_arguments()
        end = ";" if self._endline else ""
        start = "\\" if self._escape else ""
        return f"{start}{self.cmd_name}{options_str}{self._format_middle()}{argument_str}{self._format_post()}{end}"



"""
A Matplotlib-like interface for generating Tikz and Pgfplots figures
"""

__version__ = "0.1.4"

from .api import (
    gcf,
    gca,
    figure,
    ax,
    dumps,
    preview,
    save,
    xlabel,
    xticks,
    yticks,
    zticks,
    define_style,
    ylabel,
    zlabel,
    xlim,
    ylim,
    axhline,
    scale,
    node,
    fill_between,
    axvline,
    Point,
    rectangle,
    circle,
    line,
    arrow,
    plot,
    scatter,
    point,
)
from .environments.axis import Grid, Axis, AxisMode, AxisDir, View
from . import io

__all__ = [
    "gcf",
    "gca",
    "figure",
    "ax",
    "dumps",
    "preview",
    "save",
    "xlabel",
    "xticks",
    "yticks",
    "zticks",
    "define_style",
    "ylabel",
    "zlabel",
    "xlim",
    "ylim",
    "axhline",
    "scale",
    "node",
    "fill_between",
    "axvline",
    "Point",
    "rectangle",
    "circle",
    "line",
    "arrow",
    "plot",
    "scatter",
    "point",
    "Grid",
    "Axis",
    "AxisMode",
    "AxisDir",
    "View",
    "io",
]



from ..command import Command


class FillBetween(Command):

    def __init__(self, name1: str,
                 name2: str,
                 **options,
                 ):
        self.curve_name1 = name1
        self.curve_name2 = name2
        super().__init__("addplot", **options)

    # def customize_label(self, **options):
    #     """Customize the options of the inline label"""
    #     self._label_opts.set_options(**options)

    def get_code(self) -> str:
        result = f"\\addplot{self._options.format()} fill between[\n"
        result += f"of = {self.curve_name1} and {self.curve_name2},\n"
        result += "];\n"
        return result



from __future__ import annotations

from ..command import Command
from ..label import Label
from ..tikzcode import Tex
from ..formatting import format_vector
# from ..options import Options
import numpy as np


class Node(Command):

    def __init__(self,
                 label: str = "",
                 name: str | None = None,
                 position: np.ndarray | Tex | str | None = None,
                 label_loc: str | None = None,
                 axis_coords: bool = False,
                 **options,
                 ):
        self.set_position(position)
        self.axis_coords = axis_coords
        self.label = Label(label)
        if label_loc:
            self.set_label_loc(label_loc)
        # self._label_opts = Options()
        self.name = name
        super().__init__("node", label)
        self.set_options(**options)

    def customize_label(self, **options):
        """Customize the options of the inline label"""
        self.label.set_options(**options)

    # def format_label(self) -> str:

    def set_label_loc(self, loc: str):
        self.label.set_location(loc)

    def set_position(self, position: str | Tex | np.ndarray | None):
        if isinstance(position, str):
            position = Tex(position)
        if not isinstance(position, Tex) and \
                position is not None:
            position = np.array(position)
        self._position = position

    def _format_middle(self) -> str:
        if self._position is None:
            return ""
        coords = ""
        if isinstance(self._position, np.ndarray):
            coords = format_vector(self._position, separator=", ")
            if self.axis_coords:
                coords = "axis cs: " + coords
        elif isinstance(self._position, Tex):
            coords = self._position.get_code()
        else:
            raise TypeError(
                f"Got unexpected type `{type(self._position)}` for the position of a node."
            )
        middle = ""
        if self.name:
            middle += f"({self.name})"
        if coords:
            middle += f" at ({coords})"
        return middle

    def get_code(self):
        # if self._label_loc is not None:
        if len(self.label.options) > 0:
            self._arguments.clear()  # Remove the label as an argument.
            self.add_argument("")    # Replace it with the empty string, because a node needs a (potentially empty) label text.
            self.set_option("label", self.label.get_code())
        return super().get_code()



from __future__ import annotations

from ..command import Command
from .node import Node
from ..formatting import format_vector
import numpy as np
from typing import Iterable
from .connector import Connector


class Draw(Command):

    def __init__(self,
                 points: Iterable[np.ndarray | str | Node],
                 connector: Connector,
                 **options,
                 ):
        self.points = points
        self._connector = connector
        super().__init__("draw", **options)

    def _format_pt(self, pt: np.ndarray | str | Node) -> str:
        if isinstance(pt, Node):
            return f"{pt.name}"
        if isinstance(pt, (str, float)):
            return f"{pt}"
        return format_vector(pt, ", ")

    def _format_middle(self) -> str:
        formatted_pts = [f"({self._format_pt(p)})" for p in self.points]
        connector_str = self._connector.get_code()
        return connector_str.join(formatted_pts)



def format_options(**options) -> str:
    opts = []
    for name, value in options.items():
        if isinstance(value, bool):
            if value:
                opts.append(str(name))
        else:
            opts.append(f"{name}={value}")

    if opts:
        options_str = ',\n'.join(opts)
        return f"[{options_str}]"
    return ""


def wrap_env(envname: str, wrapped: str, **options) -> str:
    return f"\\begin{{{envname}}}{format_options(**options)}\n{wrapped}\n\\end{{{envname}}}"


def format_list(items):
    return ', '.join([str(item) for item in items])



from __future__ import annotations

import numpy as np
from .commands.addplot import Addplot, Addplot3d
from .constants import MAX_NUMBER


def remove_huge_nbs(arr):
    mask = np.abs(arr) > MAX_NUMBER
    arr[mask] = np.sign(arr[mask]) * np.inf
    return arr


def create_plot(x: np.ndarray, y: np.ndarray, z: np.ndarray, label: str | None = None, inline_label: bool = False, **options) -> list[Addplot]:
    plot3d = False

    if y is None:  # Plot index vs. x
        if np.ndim(x) == 0:  # Constant
            datasets = [np.array(x)]
        else:
            datasets = (
                np.hstack((np.arange(len(row))[:, np.newaxis],
                           row[:, np.newaxis]))
                for row in np.atleast_2d(x)
            )
    else:
        if z is None:
            datasets = (
                np.hstack((x_row[:, np.newaxis], y_row[:, np.newaxis]))
                for x_row, y_row in zip(np.atleast_2d(x), np.atleast_2d(y))
            )
        else:
            plot3d = True
            datasets = (
                np.hstack((x_row[:, np.newaxis], y_row[:, np.newaxis], z_row[:, np.newaxis]))
                for x_row, y_row, z_row in zip(np.atleast_2d(x), np.atleast_2d(y), np.atleast_2d(z))
            )

    def iter_label():
        if label is None:
            yield None
        elif isinstance(label, str):
            while True:
                yield label
        else:
            for lab in label:
                yield lab
            yield None

    def forget_plot(label):
        if (label is None or inline_label) \
                and ("forget_plot" not in options) \
                and ("forget plot" not in options):
            return {"forget plot": True}
        return {}

    # Set large numbers to infinity because pgfplots doesn't handle them.
    datasets = (remove_huge_nbs(a) for a in datasets)

    constructor = Addplot3d if plot3d else Addplot

    return [constructor(dataset, lab, inline_label=inline_label, **forget_plot(lab), **options) for lab, dataset in zip(iter_label(), datasets)]



from __future__ import annotations

from .draw import Draw
from .node import Node
from .connector import Connector

import numpy as np


class Circle(Draw):

    def __init__(self, point: np.ndarray | str | Node, radius: float, **options):
        self.radius = radius
        self.center = point

        super().__init__([point, str(radius)], connector=Connector("circle"), **options)

    # Override how points work
    @property
    def points(self) -> list[np.ndarray | str]:
        return [self.center, str(self.radius)]

    @points.setter
    def points(self, p):
        pass



from __future__ import annotations

import numpy as np
from .options import Options
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .label import Label


def format_options(replace_underscores=True, with_brackets=True, **options) -> str:
    """Convert the given dictionary of options to a string of options for
    a LaTeX environment.
    """
    opts = []
    for name, value in options.items():
        if isinstance(value, bool):
            if value:
                opts.append(str(name))
        elif isinstance(value, Options):
            opts.append(f"{name}={{{format_options(with_brackets=False, **value)}}}")
        else:
            if replace_underscores:
                name = name.replace("_", " ")
            opts.append(f"{name}={value}")

    if opts:
        options_str = ',\n'.join(opts)
        return f"[{options_str}]" if with_brackets else options_str
    return ""


def wrap_env(envname: str, wrapped: str, **options) -> str:
    r"""Wrap the ``wrapped`` string into a LaTeX environment with the name ``envname``.
    The dictionary of options represent options for the LaTeX environment.

    Example:

        ```
        wrap_env("axis", r"\addplot", width=4)
        ```
        generates
        ```
        \begin{axis}[width=4]
            \addplot
        \end{axis}
        ```
    """

    return f"\\begin{{{envname}}}{format_options(**options)}\n{wrapped}\n\\end{{{envname}}}"


def format_vector(vector: np.ndarray,
                  separator: str = " ",
                  ints: bool = False) -> str:
    """Format a vector as a string by joining it with the provided separator."""
    if ints:
        formatted = [f"{int(element):d}" for element in vector]
    else:
        formatted = [f"{element:2.9f}" for element in vector]
    return separator.join(formatted)


def format_matrix(matrix: np.ndarray) -> str:
    """Format a matrix for use in an \\addplot command.

    Args:
        matrix (np.ndarray): data

    Returns:
        str: Formatted table
    """
    if isinstance(matrix, str):
        return matrix
    if matrix is None:
        return ""
    if np.ndim(matrix) == 0:
        return f"{float(matrix):2.16f}"
    if np.ndim(matrix) == 1:
        return format_vector(matrix)
    elif np.ndim(matrix) == 2:
        return '\n'.join([format_vector(row) for row in matrix])
    else:
        raise ValueError(f"Formatting {np.ndim(matrix)}-dimensional data is not supported.")


def format_plot_command(
        data: np.ndarray,
        raw_options: str = "",
        suffix: str = "",
        plot3d: bool = False,
        plotplus: bool = False,
        label: "Label" | None = None,
        inline_label: bool = False,
        labelopts: str = "") -> str:
    """Return a plot command to plot the given data in pgfplots.
    """
    labelcmd = ""
    if label:
        if inline_label:
            suffix += f"node{labelopts} {{{label}}}"
        else:
            labelcmd = f"\\addlegendentry{{{label}}}"
    tablecmd = "table" if (data is not None) and (np.ndim(data) >= 1) else ""
    return f"""
\\addplot{'3' if plot3d else ''}{'+' if plotplus else ''}{raw_options}
{tablecmd}{{%
{format_matrix(data)}
}}{suffix};
{labelcmd}
"""



from .tikzcode import Tex
from .options import OptionsMixin
from itertools import product


class Label(Tex, OptionsMixin):
    """Class to represent an inline label"""

    def __init__(self, contents: str, **options):
        super().__init__(contents)
        self.init_options(**options)

    def get_code(self) -> str:
        return f"{{{self.options.format()}:{self.code}}}"

    def set_location(self, loc) -> str:
        topbot = ("", "above", "below")
        leftright = ("", "left", "right")
        allowed = [" ".join([vert, hor]).strip() for vert, hor in product(topbot, leftright)]

        if loc not in allowed:
            raise ValueError(f"Label location {loc} is not a valid position. Expected one of `{allowed}`.")
        self.set_option(loc, True)

    def contains_position_setting(self, options: dict) -> bool:
        positions = ("above", "below", "left", "right")
        return any((k.startswith(pos)) for k, pos in product(self._label_opts, positions))



from __future__ import annotations

from ..command import Command
from ..style import Style
from collections import OrderedDict


class Tikzset(Command):

    def __init__(self,
                 styles: list[Style] | None = None,
                 ):
        self._style_map = OrderedDict()
        if styles:
            for style in styles:
                self._style_map[style.name] = style

        super().__init__("tikzset")
        self._endline = False

    def add_argument(self, argument: Style):
        raise NotImplementedError("To add a new style to tikzset, use the ``set_style`` method.")

    @property
    def arguments(self) -> list[Style]:
        return list(self._style_map.values())

    def set_style(self, name: str, **options):
        if name in self._style_map:
            style: Style = self._style_map[name]
            return style.set_options(**options)
        new_style = Style(name, **options)
        self._style_map[name] = new_style

    def get_code(self) -> str:
        if not self._style_map:
            return ""
        return super().get_code()

    def remove_style(self, name: str):
        try:
            self._style_map.pop(name)
        except KeyError:
            raise UserWarning("Tried to remove undefined style.")



from __future__ import annotations

from typing import Union
from .formatting import format_options


class Tex:

    def __init__(self, code: str):
        self.code = code

    def __str__(self) -> str:
        return self.get_code()

    def get_code(self) -> str:
        return self.code


class TikzCode:
    """Low-level representation of tikz code.
    The tikz code is represented as a list of lines.
    The methods of this function can be used to easily add new
    lines to the tikz code.
    """

    def __str__(self) -> str:
        return self.get_code()

    def __repr__(self) -> str:
        return f"{self.__class__}\n{self.get_code()}"

    def __init__(self):
        self.lines = []

    def __cmd(self, cmd: str) -> "TikzCode":
        self.add_line(Tex(cmd))
        return self

    def colorlet(self, colorname: str, colordef: str) -> "TikzCode":
        """Define a new color using the ``\\colorlet`` command of xcolor."""
        return self.__cmd(f"\\colorlet{{{colorname}}}{{{colordef}}}")

    def definecolor(self, colorname: str, kind: str, definition: str) -> "TikzCode":
        """Define a new color using the ``\\definecolor`` command."""
        return self.__cmd(f"\\definecolor{{{colorname}}}{{{kind}}}{{{definition}}}")

    def usepackage(self, package: str, **options) -> "TikzCode":
        return self.__cmd(f"\\usepackage{format_options(**options)}{{{package}}}")

    def newcommand(self, command: str, definition, n_args: int = 0) -> "TikzCode":
        argcnt = f"[{n_args}]" if n_args > 0 else ""
        return self.__cmd(f"\\newcommand{{\\{command}}}{argcnt}{{{definition}}}")

    def node(self,
             label_text: str, name: str = "", location: Union[tuple, str] = "",
             **options) -> "TikzCode":
        location_str = f"at {location}" if location else ""
        return self.__cmd(f"\\node{format_options(**options)} ({name}) {location_str} {{{label_text}}};")

    def coordinate(self, name: str = "", location: Union[tuple, str] = "", **options) -> "TikzCode":
        location_str = f"at {location}" if location else ""
        return self.__cmd(f"\\coordinate{format_options(**options)} ({name}) {location_str};")

    def draw(self, *coordinates, **options) -> "TikzCode":
        return self.__cmd(f"\\draw{format_options(**options)} {'--'.join(coordinates)};")

    def add_line(self, line: Tex | str) -> "TikzCode":
        if isinstance(line, str):
            line = Tex(line)
        self.lines.append(line)
        return self

    def get_code(self) -> str:
        return "\n".join(line.get_code() for line in self.lines) + "\n"



from .node import Node
from .addplot import Addplot, Addplot3d
from .connector import Connector
from .draw import Draw
from .circle import Circle
from .fillbetween import FillBetween

__all__ = [
    "Node",
    "Addplot",
    "Addplot3d",
    "Connector",
    "Draw",
    "Circle",
    "FillBetween",
]



from dataclasses import dataclass, field
from enum import Enum
from typing import Callable, Iterator, Union, List
from string import Template
import os
import numpy as np


@dataclass
class Environment:

    def __init__(self, name: str, **options):
        self.name = name
        self.options = {name.replace("_", " "): value for name, value in options.items()}
        self.content = TikzCode()

    def add(self, content: Union['Environment', str]):
        if isinstance(content, Environment):
            content = content.get_code()
        self.content.add_line(content)

    def get_code(self) -> str:
        return wrap_env(self.name, self.content.get_code(), **self.options)


@dataclass
class View:
    azimuth: int
    elevation: int

    def __str__(self) -> str:
        return f"{{{self.azimuth}}}{{{self.elevation}}}"


class AxisMode(Enum):
    normal = "normal"
    linear = "linear"
    log = "log"


class AxisDir(Enum):
    normal = "normal"
    reverse = "reverse"


class Grid(Enum):
    minor = "minor"
    major = "major"
    both = "both"
    none = "none"


class Axis(Environment):

    def __init__(self,
                 xlabel: str = None,
                 ylabel: str = None,
                 view: View = None,
                 width: str = None,
                 height: str = None,
                 scale_only_axis: bool = False,
                 axis_x_line: str = None,
                 axis_y_line: str = None,
                 axis_z_line: str = None,
                 xmode: AxisMode = None,
                 ymode: AxisMode = None,
                 zmode: AxisMode = None,
                 x_dir: AxisDir = None,
                 y_dir: AxisDir = None,
                 z_dir: AxisDir = None,
                 axis_equal: bool = False,
                 axis_equal_image: bool = False,
                 xmin: float = None, xmax: float = None,
                 ymin: float = None, ymax: float = None,
                 zmin: float = None, zmax: float = None,
                 grid: Grid = None,
                 xminorgrids: bool = None,
                 yminorgrids: bool = None,
                 zminorgrids: bool = None,
                 xmajorgrids: bool = None,
                 ymajorgrids: bool = None,
                 zmajorgrids: bool = None,
                 xtick: str = None,
                 ytick: str = None,
                 ztick: str = None,
                 **extra
                 ):
        arguments = locals()
        arguments.pop("extra")
        arguments.pop("self")
        arguments.pop("__class__")
        arguments = {key.replace("_", " "): value for (key, value) in arguments.items()
                     if value is not None
                     }
        super().__init__("axis", **arguments, **extra)


class TikzPicture(Environment):

    def __init__(self, standalone: bool = True, **options):
        super().__init__("tikzpicture", **options)
        self.standalone = standalone
        self.preamble = TikzCode()

    def add_preamble_line(self, line: str):
        self.preamble.add_line(line)

    def get_code(self) -> str:
        preamble = ""
        if self.standalone:
            preamble += "\\documentclass[tikz]{standalone}\n"
            preamble += self.preamble.get_code()
        else:
            preamble = ""

        content = self.content.get_code()
        content = wrap_env("tikzpicture", content)
        if self.standalone:
            content = wrap_env("document", content)

        return preamble + content

        # -----------------------------------------------------------
        # IO functions
        # -----------------------------------------------------------


class ColorCycler:

    def __init__(self, cmapname: str):
        import matplotlib.pyplot as plt
        self.cmap = plt.get_cmap(cmapname)
        self.pgfdata = PlotData()
        self.__seen_colors = set()
        self.__idx = 0

    def __next__(self) -> str:
        color = self.cmap(self.__idx)
        colorname = name_from_rgb(color)
        if not colorname in self.__seen_colors:
            self.pgfdata.colordefs.append(get_colordef_command(color))
        self.__idx += 1
        return colorname

    def get_definitions(self):
        return self.pgfdata


def load_template_path(path: str) -> Template:
    """Import template with path ``path``.

    Args:
        file (str): filename of template

    Returns:
        Template: Template
    """
    with open(path, "r") as template_file:
        template = template_file.read()
    return Template(template)


def load_template(file: str) -> Template:
    """Import template with filename ``file`` from the ``.template`` directory.

    Args:
        file (str): filename of template

    Returns:
        Template: Template
    """
    template_file = os.path.join(".template", file)
    return load_template_path(template_file)


def write_output(data: str, outdir: str = ".generated_tikz", outfile: str = "plot_rationale_3d.tex") -> str:
    """Write the given data to a file ``outdir/outfile``.
    Will create ``outdir`` if it does not yet exists.

    The full output path is also returned.
    """
    os.makedirs(outdir, exist_ok=True)
    outpath = os.path.join(outdir, outfile)

    with open(outpath, "w") as f:
        f.write(data)

    return outpath


# -----------------------------------------------------------
# Tikz generation utilities
# -----------------------------------------------------------

def _preprocess_color(color: tuple) -> tuple:
    if len(color) == 4:
        if color[-1] != 1:
            raise UserWarning("RGBA is not supported. Setting alpha channel to fully opaque.")
        color = color[:-1]
    return color


def format_vector(vector: np.ndarray, separator: str = " ", ints: bool = False) -> str:
    """Format a vector as a string by joining it with the provided separator."""
    if ints:
        formatted = [f"{int(element):d}" for element in vector]
    else:
        formatted = [f"{element:2.9f}" for element in vector]
    return separator.join(formatted)


def format_matrix(matrix: np.ndarray) -> str:
    """Format a matrix for use in an \\addplot command.

    Args:
        matrix (np.ndarray): data

    Returns:
        str: Formatted table
    """
    if np.ndim(matrix) == 0:
        return f"{float(matrix):2.16f}"
    if np.ndim(matrix) == 1:
        return format_vector(matrix)
    elif np.ndim(matrix) == 2:
        return '\n'.join([format_vector(row) for row in matrix])
    else:
        raise ValueError(f"Plotting {np.ndim(matrix)}-dimensional data is not supported.")


def get_plot_command(data: np.ndarray, raw_options: str = "", suffix: str = "",
                     plot3d: bool = True, plotplus: bool = False,
                     label: str = None,
                     inline_label: bool = False,
                     labelopts: str = "") -> str:
    """Return a plot command to plot the given data in pgfplots.
    """
    labelcmd = ""
    if label:
        if inline_label:
            suffix += f"\\node[{labelopts}] {label}"
        else:
            labelcmd = f"\\addlegendentry{{{label}}}"
    tablecmd = "table" if np.ndim(data) >= 1 else ""
    return f"""
\\addplot{'3' if plot3d else ''}{'+' if plotplus else ''}[{raw_options}]
{tablecmd}{{%
{format_matrix(data)}
}}{suffix};
{labelcmd}
"""


def get_errorbar_command_raw(x_data: np.ndarray, y_data: np.ndarray, ymax_data: np.ndarray, ymin_data: np.ndarray,
                             raw_options: str = "", suffix: str = "", plotplus: bool = False,
                             label: str = None,
                             inline_label: bool = False,
                             labelopts: str = "") -> str:
    """Return a plot command with errorbars in pgfplots.
    """
    labelcmd = ""
    if label:
        if inline_label:
            suffix += f"node[{labelopts}] {{{label}}}"
        else:
            labelcmd = f"\\addlegendentry{{{label}}}"

    datarows = [f'{x:2.16f} {y:2.16f} {ym:2.16f} {yp:2.16f}' for x, y, ym, yp in zip(x_data, y_data, y_data - ymin_data, ymax_data - y_data)]
    data_fmt = "\n".join(datarows)
    return f"""
\\addplot{'+' if plotplus else ''}[{raw_options}]
plot[error bars/.cd, y dir=both, y explicit]
table[x=x,y=y, y error plus expr=\\thisrow{{ymax}},y error minus expr=\\thisrow{{ymin}}]
{{%
x y ymin ymax
{data_fmt}
}}{suffix};
{labelcmd}
"""


def get_errorbar_command(xdata: np.ndarray, ydata: np.ndarray, quantile: float,
                         raw_options: str = "", suffix: str = "", plotplus: bool = False,
                         label: str = None,
                         inline_label: bool = False,
                         labelopts: str = "") -> PlotData:
    """High level command to generate an errorbar plot in pgfplots.

    Args:
        xdata (np.ndarray): 1D array for the x-axis
        ydata (np.ndarray): 2D array for the y-axis. Error bars are computed with respect to axis 0.
        quantile (float): quantile level to compute the errorbars from.
        raw_options (str, optional): Defaults to "".
        suffix (str, optional): Defaults to "".
        plotplus (bool, optional): Defaults to False.
        label (str, optional): Label for the legend. Defaults to None.
    """
    upper_quant = np.quantile(ydata, max(quantile, 1 - quantile), axis=0)
    lower_quant = np.quantile(ydata, min(quantile, 1 - quantile), axis=0)
    means = np.mean(ydata, axis=0)

    return PlotData([get_errorbar_command_raw(xdata, means, upper_quant, lower_quant, raw_options, suffix, plotplus, label, inline_label, labelopts)])


def get_boxplot_command(data: np.ndarray, raw_options: str = "", suffix: str = "") -> str:
    """Return a plot command to plot the given data in pgfplots.
    """
    return get_plot_command(data, f"boxplot, {raw_options}", suffix, False, True)


def format_color(color: tuple) -> Iterator[str]:
    """Return an iterator over strings representing the rgb values of the given color.

    Assumes that the given color values are in the range (0,1), and outputs to (0,255),
    as used in tikz.

    Args:
        color (tuple): rgb values in range (0,1).

    Yields:
        Iterator[str]: Iterator over strings representing the color values in (0,255).
    """
    color = _preprocess_color(color)
    for c in color:
        yield str(round(c * 255))


def name_from_rgb(color: tuple) -> str:
    """Generate a name for the given rgb color.

    Args:
        color (tuple): rgb values in the (0,1)-range.

    Returns:
        str: name for the color.
    """
    return 'color' + ''.join(format_color(color))


def get_colordef_command(color: tuple) -> str:
    """Generate a command to define the given color.

    Args:
        color (tuple): rgb values in the (0,1)-range.

    Returns:
        str: definecolor-command
    """
    color = _preprocess_color(color)
    rgbcomma = ", ".join(format_color(color))
    return f"\definecolor{{{name_from_rgb(color)}}}{{RGB}}{{{rgbcomma}}}"


def get_plot_command_vector(base: np.ndarray, direction: np.ndarray, *,
                            options="", color: tuple = None, suffix: str = "") -> PlotData:
    """Generate a pgfplots command for a vector (drawn as an arrow).
    """
    start = base
    end = direction + start
    if color is not None:
        colorname = name_from_rgb(color)
        newcolor = get_colordef_command(color)
    else:
        newcolor = ""
        colorname = ""
    return PlotData(
        f"\draw[{','.join(['->, >=stealth', options, colorname])}] (axis cs: {format_vector(start, separator=', ')}) -- (axis cs: {format_vector(end, separator=', ')}) {suffix};",
        newcolor
    )


STANDALONE_SUBSTITUTIONS = dict(
    documentclass=r"\documentclass[tikz, dvipsnames]{standalone}",
    begindocument=r"\begin{document}",
    enddocument=r"\end{document}"
)
REGULAR_SUBSTITUTIONS = dict(
    documentclass="",
    begindocument="",
    enddocument=""
)


def generate_and_build(generator: Callable[[str, bool], PlotData], template_file: str, outfile: str, *,
                       copy_dst_dir: str = ""):
    """Fill the given ``template_file`` with data and write the resulting tikz-file to ``outfile``.
    If ``copy_dst_dir`` is the empty string, then a standalone document is generated which is compiled.
    Otherwise, the tex file is copied to ``copy_dst_dir``, where it can be included in another document.

    Args:
        generator (Callable[[str, bool], PlotData]): Given a path to a template file and a boolean specifying whether the plot is standalone, generate a ``PlotData`` object which contains all the plots to be generated.
        template_file (str): Path to the template file.
        outfile (str): Path to the output ``.tex`` file.
        copy_dst_dir (str, optional): Path to a directory to copy the result to. If empty, then no copying is done. Defaults to "".
    """
    import subprocess

    template = load_template(template_file)

    standalone = copy_dst_dir == ""
    if standalone:
        substitutions = STANDALONE_SUBSTITUTIONS
        preamble_data = [r"\usepackage{pgfplots}", r"\pgfplotsset{compat=1.8}", r"\usetikzlibrary{arrows,arrows.meta,3d}"]
    else:
        substitutions = REGULAR_SUBSTITUTIONS
        preamble_data = [r"\usetikzlibrary{arrows,arrows.meta,3d}"]

    plots = PlotData(preamble=preamble_data)

    plots.extend(generator(template_file, standalone))

    generated_substitutions = plots.get_substitution_dict()
    substitutions.update(generated_substitutions)
    output = template.safe_substitute(substitutions)

    # Write to file
    outfile = write_output(output, outfile=outfile)

    if standalone:
        # Build the latex file
        subprocess.call(["latexmk", "-interaction=nonstopmode", "-quiet", "-shell-escape", "-file-line-error", "-cd", "-pdf", outfile])
        print(f"Generated tikz code in \n{os.path.abspath(outfile)}")
    else:
        print(f"Generated tikz code in \n{os.path.abspath(outfile)}")
        # Copy to provided directory
        import shutil
        print(f"Copying to {os.path.join(copy_dst_dir, os.path.split(outfile)[-1])}")
        shutil.copy(os.path.abspath(outfile), copy_dst_dir)


def get_axis_settings(**kwargs) -> PlotData:
    options = []
    for key, value in kwargs.items():
        options.append(get_axis_option(key, value))
    return PlotData(extraaxisoptions=options)


def get_axis_option(key: str, value: str = None) -> str:
    if value is None:
        return f"{key},"
    return f"{key}={value},"

# -----------------------------------------------------------
# CLI
# -----------------------------------------------------------


def parse_cli():
    """Parse common cli arguments"""
    import argparse
    parser = argparse.ArgumentParser("Generate tikz code for cdc2023 paper.")
    parser.add_argument("--outdir", "-o", type=str, default="")
    args = parser.parse_args()
    args.standalone = len(args.outdir) == 0
    return args



from ..command import Command


class Connector(Command):

    def __init__(self, desc: str, **options):
        super().__init__(f" {desc} ", **options)
        self._endline = False
        self._escape = False



from __future__ import annotations

from dataclasses import dataclass
from .tikzcode import TikzCode, Tex
from .options import OptionsMixin
from typing import Union
from .formatting import wrap_env


@dataclass
class Environment(Tex, OptionsMixin):
    """Representation of a LaTeX environment."""

    def __init__(self, name: str, **options):
        self.name = name
        self.init_options(**options)
        self.content = TikzCode()
        self.requirements = TikzCode()  # Tikz code that should be added to the preamble

    def add(self, content: Union[Tex, str]):
        self.content.add_line(content)

    def get_code(self) -> str:
        return wrap_env(self.name, self.content.get_code(), **self.options)

    def add_requirement(self, req: str | Tex):
        self.requirements.add_line(req)

    def requires_package(self, package_name: str, **options):
        self.requirements.usepackage(package_name, **options)



from __future__ import annotations

from .. import formatting
from ..tikzcode import TikzCode
from ..commands.tikzset import Tikzset
# from ..command import Command
from .. import environment as env
from . import axis as ax


class TikzPicture(env.Environment):

    def __init__(self, standalone: bool = True, **options):
        super().__init__("tikzpicture", **options)
        self.standalone = standalone
        self.preamble = TikzCode()
        self._styles = Tikzset()
        # self._styles: OrderedDict[str, Options] = OrderedDict()

    def add_preamble_line(self, line: str):
        self.preamble.add_line(line)

    def remove_style(self, name: str):
        self._styles.remove_style(name)

    def set_style(self, name: str, **options):
        self._styles.set_style(name, **options)

    def define_style(self, name: str, **options):
        """Define (or update if it already exists) a style with the given name."""
        DeprecationWarning("`define_style` is deprecated. Use `set_style` instead.")
        self.set_style(name, **options)

    def add_axis(self, axis: ax.Axis | None = None):
        axis = ax.Axis() if axis is None else axis
        self.add(axis)
        self.preamble.usepackage("pgfplots")
        # self.preamble.add_line(Command("pgfplotsset", "compat=1.18"))

    def generate_preamble(self) -> str:
        preamble = self.preamble.get_code()
        for content in self.content.lines:
            if isinstance(content, env.Environment):
                preamble += (content.requirements.get_code())
        return preamble

    def format_styles(self) -> str:
        if not self._styles:
            return ""
        return self._styles.get_code() + "\n"

    def get_code(self) -> str:
        preamble = ""
        if self.standalone:
            preamble += "\\documentclass[tikz, margin=5]{standalone}\n"
            preamble += self.generate_preamble()
        else:
            preamble += "% Add the following lines in the preamble: \n"
            for line in self.generate_preamble().splitlines():
                preamble += f"%{line}\n"

        preamble += self.format_styles()
        content = self.content.get_code()
        content = formatting.wrap_env("tikzpicture", content, **self.options)
        if self.standalone:
            content = formatting.wrap_env("document", content)

        return preamble + content

    def export(self, filename: str):
        # Create the file directory if it doesn't exist already.
        import os
        parent = os.path.dirname(filename)
        if parent:
            os.makedirs(os.path.dirname(filename), exist_ok=True)

        if not filename.endswith(".tex") and \
           not filename.endswith(".tikz"):
            filename = filename + ".tex"

        with open(filename, "w") as f:
            f.write(self.get_code())

    def preview(self):
        from ..io import preview_latex_doc
        is_standalone = self.standalone
        self.standalone = True
        prev = preview_latex_doc(self.get_code())
        self.standalone = is_standalone  # Reset to initial state
        return prev



from __future__ import annotations

from ..command import Command
from ..options import Options
from ..formatting import format_plot_command
import numpy as np


class __AddplotBase(Command):

    def __init__(self, data: np.ndarray,
                 plot3d: bool,
                 label: str,
                 inline_label: bool,
                 **options,
                 ):

        self.data = data
        self.label = label
        self._plot3d = plot3d
        self._inline_label = inline_label
        self._label_opts = Options(
            pos=0.7, above=True
        )
        super().__init__("addplot", **options)

    def customize_label(self, **options):
        """Customize the options of the inline label"""
        self._label_opts.set_options(**options)

    def get_code(self) -> str:
        return format_plot_command(self.data,
                                   raw_options=self.options.format(),
                                   plot3d=self._plot3d,
                                   label=self.label,
                                   inline_label=self._inline_label,
                                   labelopts=self._label_opts.format(),
                                   plotplus=False
                                   )


class Addplot(__AddplotBase):

    def __init__(self, data: np.ndarray,
                 label: str | None = None,
                 inline_label: bool = False,
                 **options,
                 ):
        super().__init__(data, False, label, inline_label, **options)


class Addplot3d(__AddplotBase):

    def __init__(self, data: np.ndarray,
                 label: str | None = None,
                 inline_label: bool = False,
                 **options,
                 ):
        super().__init__(data, True, label, inline_label, **options)



from collections.abc import Mapping
from collections import OrderedDict


class Options(Mapping):

    def __init__(self, **options):
        self._opt_dict = OrderedDict(**{key.replace("_", " "): value
                                        for (key, value) in options.items()
                                        if value is not None})

    def format(self, include_brackets: bool = True) -> str:
        from .formatting import format_options
        return format_options(with_brackets=include_brackets, **self._opt_dict)

    def set_option(self, name: str, value: str):
        self._opt_dict[name.replace("_", " ")] = value

    def set_options(self, **options):
        for k, v in options.items():
            self.set_option(k, v)

    # Dict-like methods

    def __getitem__(self, key):
        return self._opt_dict[key]

    def __setitem__(self, key, value):
        self._opt_dict[key] = value

    def __delitem__(self, key):
        del self._opt_dict[key]

    def __iter__(self):
        return iter(self._opt_dict)

    def __len__(self):
        return len(self._opt_dict)

    def __dict__(self):
        return self._opt_dict


class OptionsMixin:

    @property
    def options(self) -> Options:
        return self._options

    def init_options(self, **options):
        self._options = Options(**options)

    def set_option(self, key: str, value: str):
        self._options.set_option(key, value)

    def set_options(self, **options):
        self._options.set_options(**options)

    def update_option(self, name: str, inner_key: str, value: str):
        suboption = self.options.get(name, Options())
        if not isinstance(suboption, Options):
            raise TypeError(f"Cannot update sub-options. Option {suboption} is a string.")
        suboption.set_option(inner_key, value)
        self.options[name] = suboption



from .tikzcode import Tex
from .options import OptionsMixin
from .formatting import format_options


class Style(Tex, OptionsMixin):

    def __init__(self, name: str, **options):
        self.name = name
        self.init_options(**options)

    def get_code(self) -> str:
        return f"{self.name}/.style={{ {format_options(with_brackets=False, **self.options)} }}"



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from ..environment import Environment
from ..util import format_list


@dataclass
class View:
    azimuth: int
    elevation: int

    def __str__(self) -> str:
        return f"{{{self.azimuth}}}{{{self.elevation}}}"


class AxisMode(Enum):
    normal = "normal"
    linear = "linear"
    log = "log"


class AxisDir(Enum):
    normal = "normal"
    reverse = "reverse"


class Grid(Enum):
    minor = "minor"
    major = "major"
    both = "both"
    none = "none"


class Axis(Environment):

    def __init__(self,
                 xlabel: str | None = None,
                 ylabel: str | None = None,
                 view: View | None = None,
                 width: str | None = None,
                 height: str | None = None,
                 scale_only_axis: bool = False,
                 axis_x_line: str | None = None,
                 axis_y_line: str | None = None,
                 axis_z_line: str | None = None,
                 xmode: AxisMode | None = None,
                 ymode: AxisMode | None = None,
                 zmode: AxisMode | None = None,
                 x_dir: AxisDir | None = None,
                 y_dir: AxisDir | None = None,
                 z_dir: AxisDir | None = None,
                 axis_equal: bool = False,
                 axis_equal_image: bool = False,
                 xmin: float | None = None, xmax: float | None = None,
                 ymin: float | None = None, ymax: float | None = None,
                 zmin: float | None = None, zmax: float | None = None,
                 grid: Grid | None = None,
                 xminorgrids: bool | None = None,
                 yminorgrids: bool | None = None,
                 zminorgrids: bool | None = None,
                 xmajorgrids: bool | None = None,
                 ymajorgrids: bool | None = None,
                 zmajorgrids: bool | None = None,
                 xtick: str | None = None,
                 ytick: str | None = None,
                 ztick: str | None = None,
                 **extra
                 ):
        arguments = locals()
        arguments.pop("extra")
        arguments.pop("self")
        arguments.pop("__class__")
        arguments = {key.replace("_", " "): value for (key, value) in arguments.items()
                     if value is not None
                     }
        super().__init__("axis", **arguments, **extra)

    def set_xlabel(self, label: str):
        self.set_option("xlabel", label)

    def enlarge_limits(self, amount: float = 0.05, direction: str | None = None):
        """Enlarge the range of axes (only x or y if `direction` is provided).

        Adds the `enlarge x limits` option.

        Arguments:
            - amount: fraction by which to increase the size of the axes. The final range will be `(1 + amount) * original_length`
            - direction: "x" | "y". If provided, only apply to the corresponding axis.
        """
        allowed_dirs = ["x", "y", None]
        if direction not in allowed_dirs:
            raise ValueError(f"Expected direction to belong to {allowed_dirs}. Got {direction}")

        if direction:
            self.set_option(f"enlarge {direction} limits", amount)
        else:
            self.set_option("enlarge x limits", amount)
            self.set_option("enlarge y limits", amount)

    # def extend_axis_lines(self, amount: int | str = "10pt"):
    #     self.set_option("axis line style", Options(
    #         **{"shorten >": "-10pt", "shorten <": "-10pt"}
    #     ))

    def set_ylabel(self, label: str):
        self.options.set_option("ylabel", label)

    def set_zlabel(self, label: str):
        self.options.set_option("zlabel", label)

    def set_xlims(self, lims: tuple[float]):
        (self.options["xmin"], self.options["xmax"]) = lims

    def set_ylims(self, lims: tuple[float]):
        (self.options["ymin"], self.options["ymax"]) = lims

    def set_zlims(self, lims: tuple[float]):
        (self.options["zmin"], self.options["zmax"]) = lims

    def get_ylims(self) -> tuple:
        return (self.options.get("ymin", -1), self.options.get("ymax", 1))

    def center(self):
        """Center the axis lines"""
        self.set_option("axis lines", "center")

    def set_axis_label_position(self, axis: str, position: str):
        """Set label position of the given axis relative to the arrowhead.
        As a side-effect, it also centers the axes.

        Arguments:
            - axis [str]: "x", "y" or "z"
            - positions [str]: "left", "right", "above", "below"
        """
        axis = axis.lower()
        position = position.lower()
        known_axes = ("x", "y", "z")
        known_positions = ("left", "right", "above", "below")

        if axis not in known_axes:
            raise ValueError(f"Expected axis in {known_axes}. Got {axis}.")

        if position not in known_positions:
            raise ValueError(f"Expected position in {known_positions}. Got {position}.")

        self.center()
        anchors = dict(
            above="south",
            below="north",
            left="east",
            right="west"
        )

        self.update_option(f"{axis} label style", "anchor", anchors[position])

    def boxed(self):
        self.set_option("axis lines", "box")

    def _set_ticks(self, values: tuple[float], labels: tuple[str] | None, direction: str):
        self.set_option(f"{direction}tick", f"{{{format_list(values)}}}")
        if labels is not None:
            self.set_option(f"{direction}ticklabels", f"{{{format_list(labels)}}}")

    def set_yticks(self, values: tuple[float], labels: tuple[str] | None = None):
        self._set_ticks(values, labels, "y")

    def set_xticks(self, values: tuple[float], labels: tuple[str] | None = None):
        self._set_ticks(values, labels, "x")

    def set_zticks(self, values: tuple[float], labels: tuple[str] | None = None):
        self._set_ticks(values, labels, "z")

    def set_xmax(self, v):
        self.options["xmax"] = v

    def set_ymax(self, v):
        self.options["ymax"] = v

    def set_zmax(self, v):
        self.options["zmax"] = v

    def set_xmin(self, v):
        self.options["xmin"] = v

    def set_ymin(self, v):
        self.options["ymin"] = v

    def set_zmin(self, v):
        self.options["zmin"] = v
